from .websocket_base import WSBase, WSClientConnectionClosedException, WSClientException


class WSClient(WSBase):
    from .channels import (
        candles,
        candles_async,
        candles_unsubscribe,
        candles_unsubscribe_async,
        heartbeats,
        heartbeats_async,
        heartbeats_unsubscribe,
        heartbeats_unsubscribe_async,
        level2,
        level2_async,
        level2_unsubscribe,
        level2_unsubscribe_async,
        market_trades,
        market_trades_async,
        market_trades_unsubscribe,
        market_trades_unsubscribe_async,
        status,
        status_async,
        status_unsubscribe,
        status_unsubscribe_async,
        ticker,
        ticker_async,
        ticker_batch,
        ticker_batch_async,
        ticker_batch_unsubscribe,
        ticker_batch_unsubscribe_async,
        ticker_unsubscribe,
        ticker_unsubscribe_async,
        user,
        user_async,
        user_unsubscribe,
        user_unsubscribe_async,
    )
